@@ -1,22 +1,22 @@
 /*
- * testing-extensin.c
+ * testing-extension.c
  * This file is part of libpeas
  *
- * Copyright (C) 2011 - Garrett Regier
+ * Copyright (C) 2011-2014 - Garrett Regier
  *
- *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU Library General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
+ * libpeas is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation; either
+ * version 2.1 of the License, or (at your option) any later version.
  *
- *  This program is distributed in the hope that it will be useful,
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU Library General Public License for more details.
+ * libpeas is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
  *
- *  You should have received a copy of the GNU Library General Public License
- *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.
  */
 
 #ifdef HAVE_CONFIG_H
@@ -37,7 +37,6 @@
 #include "introspection-callable.h"
 #include "introspection-has-missing-prerequisite.h"
 #include "introspection-has-prerequisite.h"
-#include "introspection-properties.h"
 #include "introspection-unimplemented.h"
 
 typedef struct _TestFixture TestFixture;
@@ -47,6 +46,7 @@ struct _TestFixture {
   PeasPluginInfo *info;
 };
 
+static gchar *loader = NULL;
 static gchar *extension_plugin = NULL;
 
 static void
@@ -54,6 +54,7 @@ test_setup (TestFixture    *fixture,
             gconstpointer  data)
 {
   fixture->engine = testing_engine_new ();
+  peas_engine_enable_loader (fixture->engine, loader);
   fixture->info = peas_engine_get_plugin_info (fixture->engine,
                                                extension_plugin);
 }
@@ -172,9 +173,9 @@ test_extension_create_invalid (PeasEngine     *engine,
 
   /* This cannot be tested in PyGI and Seed's log handler messes this up */
   if (g_strcmp0 (extension_plugin, "extension-c") != 0 &&
+      g_strcmp0 (extension_plugin, "extension-lua51") != 0 &&
       g_strcmp0 (extension_plugin, "extension-python") != 0 &&
-      g_strcmp0 (extension_plugin, "extension-python3") != 0 &&
-      g_strcmp0 (extension_plugin, "extension-seed") != 0)
+      g_strcmp0 (extension_plugin, "extension-python3") != 0)
     {
       testing_util_push_log_hook ("*cannot add *IntrospectionHasMissingPrerequisite* "
                                   "which does not conform to *IntrospectionCallable*");
@@ -268,6 +269,113 @@ test_extension_get_settings (PeasEngine     *engine,
   g_object_unref (extension);
 }
 
+static gint
+run_in_multiple_threads (GFunc    func,
+                         gpointer user_data)
+{
+  gint i, n_threads;
+  GThreadPool *pool;
+  GError *error = NULL;
+
+  /* Avoid too many threads, but try to get some good contention */
+  n_threads = g_get_num_processors () + 2;
+
+  pool = g_thread_pool_new (func, user_data, n_threads, TRUE, &error);
+  g_assert_no_error (error);
+
+  for (i = 0; i < n_threads; ++i)
+    {
+      /* Cannot supply NULL as the data... */
+      g_thread_pool_push (pool, GUINT_TO_POINTER (i + 1), &error);
+      g_assert_no_error (error);
+    }
+
+  g_thread_pool_free (pool, FALSE, TRUE);
+
+  return n_threads;
+}
+
+static void
+multiple_threads_loaders_in_thread (guint    nth_thread,
+                                    gboolean use_nonglobal_loaders)
+{
+  gint i, j;
+  PeasEngine *engine;
+  PeasPluginInfo *info;
+  GObject *extension;
+
+  engine = testing_engine_new_full (use_nonglobal_loaders);
+  peas_engine_enable_loader (engine, loader);
+
+  info = peas_engine_get_plugin_info (engine, extension_plugin);
+  g_assert (info != NULL);
+
+  for (i = 0; i < 10; ++i)
+    {
+      g_assert (peas_engine_load_plugin (engine, info));
+
+      for (j = 0; j < 50; ++j)
+        {
+          extension = peas_engine_create_extension (engine, info,
+                                                    INTROSPECTION_TYPE_BASE,
+                                                    NULL);
+
+          g_assert (extension != NULL);
+          g_object_unref (extension);
+        }
+
+      g_assert (peas_engine_unload_plugin (engine, info));
+    }
+
+  testing_engine_free (engine);
+}
+
+static void
+test_extension_multiple_threads_global_loaders (PeasEngine     *engine,
+                                                PeasPluginInfo *info)
+{
+  run_in_multiple_threads ((GFunc) multiple_threads_loaders_in_thread,
+                           GINT_TO_POINTER (FALSE));
+}
+
+static void
+test_extension_multiple_threads_nonglobal_loaders (PeasEngine     *engine,
+                                                   PeasPluginInfo *info)
+{
+  run_in_multiple_threads ((GFunc) multiple_threads_loaders_in_thread,
+                           GINT_TO_POINTER (TRUE));
+}
+
+static void
+multiple_threads_callbacks_in_thread (guint            nth_thread,
+                                      PeasActivatable *activatable)
+{
+  gint i;
+
+  for (i = 0; i < 100; ++i)
+    peas_activatable_update_state (activatable);
+}
+
+static void
+test_extension_multiple_threads_callbacks (PeasEngine     *engine,
+                                           PeasPluginInfo *info)
+{
+  PeasExtension *extension;
+  gint n_threads, update_count;
+
+  extension = peas_engine_create_extension (engine, info,
+                                            PEAS_TYPE_ACTIVATABLE,
+                                            NULL);
+
+  n_threads = run_in_multiple_threads ((GFunc) multiple_threads_callbacks_in_thread,
+                                       extension);
+
+  g_object_get (extension, "update-count", &update_count, NULL);
+  g_assert_cmpint (update_count, ==, n_threads * 100);
+
+  g_object_unref (extension);
+}
+
 static void
 test_extension_call_no_args (PeasEngine     *engine,
                              PeasPluginInfo *info)
@@ -293,7 +401,7 @@ test_extension_call_with_return (PeasEngine     *engine,
 {
   PeasExtension *extension;
   IntrospectionCallable *callable;
-  const gchar *return_val = NULL;
+  gchar *return_val = NULL;
 
   extension = peas_engine_create_extension (engine, info,
                                             INTROSPECTION_TYPE_CALLABLE,
@@ -304,11 +412,13 @@ test_extension_call_with_return (PeasEngine     *engine,
   g_assert (peas_extension_call (extension, "call_with_return", &return_val));
   g_assert_cmpstr (return_val, ==, "Hello, World!");
 
+  g_free (return_val);
   return_val = NULL;
 
   return_val = introspection_callable_call_with_return (callable);
   g_assert_cmpstr (return_val, ==, "Hello, World!");
 
+  g_free (return_val);
   g_object_unref (extension);
 }
 
@@ -374,101 +484,6 @@ test_extension_call_multi_args (PeasEngine     *engine,
   g_object_unref (extension);
 }
 
-static void
-test_extension_properties_construct_only (PeasEngine     *engine,
-                                          PeasPluginInfo *info)
-{
-  PeasExtension *extension;
-  gchar *construct_only;
-
-  extension = peas_engine_create_extension (engine, info,
-                                            INTROSPECTION_TYPE_PROPERTIES,
-                                            "construct-only", "my-construct-only",
-                                            NULL);
-
-  g_object_get (extension, "construct-only", &construct_only, NULL);
-  g_assert_cmpstr (construct_only, ==, "my-construct-only");
-  g_free (construct_only);
-
-  g_object_unref (extension);
-}
-
-static void
-test_extension_properties_read_only (PeasEngine     *engine,
-                                     PeasPluginInfo *info)
-{
-  PeasExtension *extension;
-  gchar *read_only;
-
-  extension = peas_engine_create_extension (engine, info,
-                                            INTROSPECTION_TYPE_PROPERTIES,
-                                            NULL);
-
-  g_object_get (extension, "read-only", &read_only, NULL);
-  g_assert_cmpstr (read_only, ==, "read-only");
-  g_free (read_only);
-
-  g_object_unref (extension);
-}
-
-static void
-test_extension_properties_write_only (PeasEngine     *engine,
-                                      PeasPluginInfo *info)
-{
-  PeasExtension *extension;
-
-  extension = peas_engine_create_extension (engine, info,
-                                            INTROSPECTION_TYPE_PROPERTIES,
-                                            NULL);
-
-  g_object_set (extension, "write-only", "my-write-only", NULL);
-
-  g_object_unref (extension);
-}
-
-static void
-test_extension_properties_readwrite (PeasEngine     *engine,
-                                     PeasPluginInfo *info)
-{
-  PeasExtension *extension;
-  gchar *readwrite;
-
-  extension = peas_engine_create_extension (engine, info,
-                                            INTROSPECTION_TYPE_PROPERTIES,
-                                            NULL);
-
-  g_object_get (extension, "readwrite", &readwrite, NULL);
-  g_assert_cmpstr (readwrite, ==, "readwrite");
-  g_free (readwrite);
-
-  g_object_set (extension, "readwrite", "my-readwrite", NULL);
-
-  g_object_get (extension, "readwrite", &readwrite, NULL);
-  g_assert_cmpstr (readwrite, ==, "my-readwrite");
-  g_free (readwrite);
-
-  g_object_unref (extension);
-}
-
-static void
-test_extension_properties_prerequisite (PeasEngine     *engine,
-                                        PeasPluginInfo *info)
-{
-  PeasExtension *extension;
-  gchar *prerequisite;
-
-  extension = peas_engine_create_extension (engine, info,
-                                            INTROSPECTION_TYPE_PROPERTIES,
-                                            "prerequisite", "prerequisite",
-                                            NULL);
-
-  g_object_get (extension, "prerequisite", &prerequisite, NULL);
-  g_assert_cmpstr (prerequisite, ==, "prerequisite");
-  g_free (prerequisite);
-
-  g_object_unref (extension);
-}
-
 #define _EXTENSION_TEST(loader, path, ftest) \
   G_STMT_START { \
     gchar *full_path = g_strdup_printf (EXTENSION_TEST_NAME (%s, "%s"), \
@@ -482,11 +497,23 @@ test_extension_properties_prerequisite (PeasEngine     *engine,
   } G_STMT_END
 
 void
-testing_extension_basic (const gchar *loader)
+testing_extension_basic (const gchar *loader_)
 {
+  gint i, j;
+  gchar *loader_name;
   PeasEngine *engine;
 
-  extension_plugin = g_strdup_printf ("extension-%s", loader);
+  loader = g_strdup (loader_);
+
+  loader_name = g_new0 (gchar, strlen (loader) + 1);
+  for (i = 0, j = 0; loader[i] != '\0'; ++i)
+    {
+      if (loader[i] != '.')
+        loader_name[j++] = loader[i];
+    }
+
+  extension_plugin = g_strdup_printf ("extension-%s", loader_name);
+  g_free (loader_name);
 
   engine = testing_engine_new ();
   peas_engine_enable_loader (engine, loader);
@@ -510,6 +537,18 @@ testing_extension_basic (const gchar *loader)
 
   _EXTENSION_TEST (loader, "plugin-info", plugin_info);
   _EXTENSION_TEST (loader, "get-settings", get_settings);
+
+  _EXTENSION_TEST (loader, "multiple-threads/global-loaders",
+                   multiple_threads_global_loaders);
+  _EXTENSION_TEST (loader, "multiple-threads/nonglobal-loaders",
+                   multiple_threads_nonglobal_loaders);
+
+  /* Not needed for C plugins as they are independent of libpeas */
+  if (g_strcmp0 (loader, "c") != 0)
+    {
+      _EXTENSION_TEST (loader, "multiple-threads/callbacks",
+                       multiple_threads_callbacks);
+    }
 }
 
 void
@@ -521,16 +560,6 @@ testing_extension_callable (const gchar *loader)
   _EXTENSION_TEST (loader, "call-multi-args", call_multi_args);
 }
 
-void
-testing_extension_properties (const gchar *loader)
-{
-  _EXTENSION_TEST (loader, "properties-construct-only", properties_construct_only);
-  _EXTENSION_TEST (loader, "properties-read-only", properties_read_only);
-  _EXTENSION_TEST (loader, "properties-write-only", properties_write_only);
-  _EXTENSION_TEST (loader, "properties-readwrite", properties_readwrite);
-  _EXTENSION_TEST (loader, "properties-prerequisite", properties_prerequisite);
-}
-
 void
 testing_extension_add (const gchar   *path,
                        GTestDataFunc  func)
@@ -547,6 +576,7 @@ testing_extension_run_tests (void)
   retval = testing_run_tests ();
 
   g_free (extension_plugin);
+  g_free (loader);
 
   return retval;
 }
